/*
 * Copyright (c) 2015, Majenko Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * * Neither the name of Majenko Technologies nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

 /* well, the above wrote the wifi stack, I wrote the rest and it's GPL-V2, C Doug Coulter 2016 */

 
 //@@@ for later use:  ESP.restart() is software reset, but doesn't work on serial jig - only later on.

#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
// may not need below
//#include <ESP8266mDNS.h>
#include <WiFiUdp.h>

// pin defines as this might get ported to another platform
#define DEBUG_PIN 16
#define PULSE_PIN 12


// scheduling stuff
bool debug; // if pin is high, we're not running on the dev box
bool did_something; // we did something this loop, so hold off till next one
bool do_tellem;
bool ten_sec;

unsigned long loop_millis;
#define TENSEC_MS 10000
unsigned long tensec_z;
#define MINUTE_MS 60000
unsigned long minute_z; // milliseconds since last tell

// tellem stuff
WiFiUDP tport;
char packetBuffer[255];
#define TELLEM_PORT 53831
IPAddress myip; // my ip as assigned by dhcp
IPAddress bcip; // broadcast addr for this network
unsigned char mac[6]; // my mac, I hope
char tellem_string[40]; // What we'll broadcast
unsigned long tellems;
unsigned long errors;


// WAP stuff - uh-oh hardcoded creds @@@
const char *ssid = "machswap"; //machswap
const char *password = "fusordoug";

ESP8266WebServer server ( 80 );

// app stuff
volatile unsigned int true_time; // yep, need volatile here
volatile unsigned int false_time;
unsigned long pulse_z;
unsigned int pulse_wait;
bool pulse_state;
bool pulse_enable;

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
void init_ram()
{ // make ram clean
  debug = did_something = do_tellem = ten_sec = false;
  tellems = errors = 0;
  minute_z = tensec_z = pulse_z = millis();
  true_time = false_time = 2; // ms per phase of pulse
  pulse_wait = false_time;
  pulse_state = pulse_enable = false;
}
////////////////////////////////////
void build_tellem_string ()
{
  char i;
  String mystring;

//@@@ class String can't 0 pad hex...
// but sprintf and strncpy are lots more lines of code (less rom though)
  mystring = myip.toString();
  mystring += " ESP_";
  for (i=3;i<6;i++) 
  {
    if (16 > mac[i]) mystring += "0"; // yes, it is that dumb    
    mystring += String(mac[i],HEX);
   }
  mystring.toUpperCase(); // because damn, that's what it is on the net for some retarded opsys
  mystring.toCharArray(tellem_string, 39);
  
  Serial.println(tellem_string); // debug
}
////////////////////////////////////
void tellem()
{
  tport.beginPacket(bcip,TELLEM_PORT);
  tport.write(tellem_string);
  tport.endPacket(); 
  tellems++; // for grins, count them
  do_tellem = false; // we did it
}
////////////////////////////////////
void ditch_incoming()
{ // ignore other broadcasts, maybe stop possible memory leak.
  int packetSize = tport.parsePacket();
  if (packetSize) {
    int len = tport.read(packetBuffer, 254);
    if (debug)
    { // but hey, make sure it works if we are in debug mode!
     if (len > 0) packetBuffer[len] = 0;
     Serial.println(packetBuffer);
    }  
  }
}
////////////////////////////////////
////////////////////////////////////

//    <p><a href='/controls'>Controls</a></p>\

////////////////////////////////////
void handleRoot() {
  int i,j;

  if (j = server.args()) // will only be true for post-submit, not refresh 
  { //     due to hidden field will always be true in that case.  Get input values.
   pulse_enable = false; // we don't get anything if it is so assume
   for (i=0;i<j;i++) // scan all input
   { // there's probably a better way than compare everything to everything here?
    if (server.argName(i).equals("enable")) pulse_enable = true;
    if (server.argName(i).equals("pulsehigh")) true_time = server.arg(i).toInt();
    if (server.argName(i).equals("pulselow")) false_time = server.arg(i).toInt();
   }
  }
  if (true_time <= 0) true_time =1; // malformed input
  if (false_time <= 0) false_time =1; // malformed input
  
  String page = \
"<html>\  
  <head>\
    <title>Pulse Control</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
    <body>\
    <h1>Pulse Control</h1><p> </p>\
    <FORM action=\"/\" method=\"post\"> <INPUT type=\"hidden\" name=\"user\" value=\"1\"> ";

  page += "<INPUT type=\"text\" size=\"5\" maxlength=\"5\" name=\"pulsehigh\" value=\"";
  page += String(true_time,DEC);
  page += "\"> Pulse High, mS <br>";
  page += "<INPUT type=\"text\" size=\"5\" maxlength=\"5\" name=\"pulselow\" value=\"";
  page += String (false_time,DEC);
  page += "\"> Pulse Low, mS <br>";

// programmatic content here

  if(pulse_enable) page += "<INPUT type=\"checkbox\" name=\"enable\" value=\"1\" checked>Enable<br>";
  else page += "<INPUT type=\"checkbox\" name=\"enable\" value=\"1\" >Enable<br>";

  page += \  
"<p></p><INPUT type=\"submit\" value=\"Do it\"></FORM>\ 
    <p> </p><a href='/'>Home</a></p>\
  </body>\
</html>";  
  did_something = true; // we did something that takes time
	server.send ( 200, "text/html", page );

  if (debug)
  {
   Serial.print("true time: ");
   Serial.print(true_time);
   Serial.print(" false time: ");
   Serial.print(false_time);
   Serial.print(" enable: ");
   Serial.println (pulse_enable);
     
  }//*/ // that was painful
}
/////////////////////////////////////////////////

////////////////////////////////////////////
void handleNotFound() {
	String message = "File Not Found\n\n";
	message += "URI: ";
	message += server.uri();
	message += "\nMethod: ";
	message += ( server.method() == HTTP_GET ) ? "GET" : "POST";
	message += "\nArguments: ";
	message += server.args();
	message += "\n";

	for ( uint8_t i = 0; i < server.args(); i++ ) {
		message += " " + server.argName ( i ) + ": " + server.arg ( i ) + "\n";
	}

	server.send ( 404, "text/plain", message );
  did_something = true;
}
// *****************************************************************************/
void setup ( void ) {

   init_ram();
// decide on debug mode here
   pinMode (DEBUG_PIN,INPUT);
   debug = digitalRead(DEBUG_PIN);
   pinMode (PULSE_PIN,OUTPUT); // doh

   Serial.begin ( 115200 );
   WiFi.begin ( ssid, password );Serial.println ( "" );

	// Wait for connection
	while ( WiFi.status() != WL_CONNECTED ) {
		delay ( 500 );
		Serial.print ( "." );
	}
   /// tellem stuff
   WiFi.macAddress(mac); // get my MAC addr, as it's part of hostname
   myip = WiFi.localIP(); // class and union of dword and 4 bytes
   bcip = myip; // copy net address but
   bcip[3] = 255; // make it broadcast address

	 Serial.println ( "" );
	 Serial.print ( "Connected to " );
	 Serial.println ( ssid );
	 Serial.print ( "My IP address: " );
	 Serial.println ( WiFi.localIP() );
   Serial.print ("Signal Strength: ");
   Serial.println (WiFi.RSSI());
   server.on ( "/", handleRoot );
	 server.onNotFound ( handleNotFound );
	 server.begin();
	 Serial.println ( "HTTP server started" );
 // tellem stuff
   build_tellem_string();
   tport.begin(TELLEM_PORT); // for tellem
}
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
void loop ( void ) {
  did_something = false; // because we haven't yet on this round
	server.handleClient(); // if any hits should set did_something to true
  loop_millis = millis(); // just do this once
  debug = digitalRead(DEBUG_PIN); // semi-realtime update of debug

// set scheduling flags for various intervals
// add setting various bools to these to get something to run (eventually)
 
  if (loop_millis - minute_z >= MINUTE_MS)
  {
    minute_z = loop_millis; // update for next time
    do_tellem = true; // schedule 1/minute stuff
  }
  if (loop_millis - tensec_z >= TENSEC_MS)
  {
   tensec_z = loop_millis; 
   ten_sec = true;    
  }

// Do stuff.  Not 1::1 with how often requested on purpose - important stuff gets priority regardless.
// (because we check/do that first)  

  if (pulse_enable)
  {
    
   if (loop_millis - pulse_z >= pulse_wait)
   { // state switch
    pulse_z = loop_millis; // update time mark
    if(pulse_state) pulse_wait = false_time; // toggle what we'll wait for next
    else pulse_wait = true_time;
    
    pulse_state = !pulse_state; // toggle output state
   }
    
  } else pulse_state = false;
  digitalWrite(PULSE_PIN,pulse_state); // do it
    
// If we've already "done something" this loop, wait for next roundy round.

  if (do_tellem  && !did_something) // only if it is time AND we didn't do something else first 
  {
   tellem();
   did_something = true; // skip other things this loop
  } // every minute or so
 
   if (!did_something) ditch_incoming(); // in case it matters?
// fancy opsys, eh?
}
//********************************************************************************************

